IDENTITY_URL_PATH = 'https://api-finder.eircode.ie/Latest/findergetidentity'
EIRCODE_FINDER_URL_PATH = 'https://api-finder.eircode.ie/Latest/finderfindaddress'

ROUTING_KEYS_TOWNS_MAP = {
    'A41': ['Ballyboughal'],
    'A42': ['Garristown'],
    'A45': ['Oldtown'],
    'A63': ['Greystones'],
    'A67': ['Wicklow'],
    'A75': ['Castleblayney'],
    'A81': ['Carrickmacross'],
    'A82': ['Kells', 'Kingscourt', 'Virginia'],
    'A83': ['Enfield', 'Summerhill', 'Rathmoylon'],
    'A84': ['Ashbourne'],
    'A85': ['Dunshaughlin'],
    'A86': ['Dunboyne'],
    'A91': ['Dundalk'],
    'A92': ['Drogheda', 'Ardee', 'Dunleer'],
    'A94': ['Blackrock', 'Monkstown', 'Booterstown', 'Stillorgan'],
    'A96': ['Dun Laoghaire', 'Dalkey', 'Sallynoggin', 'Glenageary'],
    'A98': ['Bray', 'Kilmacanogue', 'Roundwood'],
    'C15': ['Trim', 'Navan', 'Athboy'],
    'D01': ['Dublin 1'],
    'D02': ['Dublin 2'],
    'D03': ['Dublin 3'],
    'D04': ['Dublin 4'],
    'D05': ['Dublin 5'],
    'D06': ['Dublin 6'],
    'D6W': ['Dublin 6W'],
    'D07': ['Dublin 7'],
    'D08': ['Dublin 8'],
    'D09': ['Dublin 9'],
    'D10': ['Dublin 10'],
    'D11': ['Dublin 11'],
    'D12': ['Dublin 12'],
    'D13': ['Dublin 13'],
    'D14': ['Dublin 14'],
    'D15': ['Dublin 15'],
    'D16': ['Dublin 16'],
    'D17': ['Dublin 17'],
    'D18': ['Dublin 18'],
    'D20': ['Dublin 20'],
    'D22': ['Dublin 22'],
    'D24': ['Dublin 24'],
    'E21': ['Cahir'],
    'E25': ['Cashel'],
    'E32': ['Carrick on Suir'],
    'E34': ['Tipperary'],
    'E41': ['Thurles', 'Templemore'],
    'E45': ['Nenagh'],
    'E53': ['Roscrea'],
    'E91': ['Clonmel', 'Fethard'],
    'F12': ['Knock', 'Swinford', 'Claremorris'],
    'F23': ['Castlebar'],
    'F26': ['Ballina', 'Bangor'],
    'F28': ['Westport', 'Achill'],
    'F31': ['Ballinrobe'],
    'F35': ['Ballyhaunis'],
    'F42': ['Roscommon'],
    'F45': ['Castlerea', 'Glenamaddy'],
    'F52': ['Boyle'],
    'F56': ['Ballymote'],
    'F91': ['Sligo', 'Manorhamilton', 'Tubbercurry'],
    'F92': ['Letterkenny'],
    'F93': ['Lifford', 'Carndonagh'],
    'F94': ['Bundoran', 'Donegal'],
    'H12': ['Cavan'],
    'H14': ['Belturbet'],
    'H16': ['Cootehill'],
    'H18': ['Monaghan'],
    'H23': ['Clones'],
    'H53': ['Ballinasloe', 'Portumna'],
    'H54': ['Tuam'],
    'H62': ['Loughrea'],
    'H65': ['Athenry'],
    'H71': ['Clifden'],
    'H91': ['Galway', 'Spiddal', 'Headford', 'Ballyvaughan'],
    'K32': ['Balbriggan'],
    'K34': ['Skerries'],
    'K36': ['Malahide', 'Donabate'],
    'K45': ['Lusk'],
    'K56': ['Rush'],
    'K67': ['Swords'],
    'K78': ['Lucan'],
    'N37': ['Athlone', 'Moate'],
    'N39': ['Longford', 'Ballymahon', 'Granard', 'Lanesborough'],
    'N41': ['Carrick On Shannon'],
    'N91': ['Mullingar', 'Kinnegad'],
    'P12': ['Macroom'],
    'P14': ['Crookstown'],
    'P17': ['Kinsale'],
    'P24': ['Cobh'],
    'P25': ['Midleton'],
    'P31': ['Ballincollig'],
    'P32': ['Rylane'],
    'P36': ['Youghal'],
    'P43': ['Carrigaline'],
    'P47': ['Dunmanway'],
    'P51': ['Mallow', 'Rathmore', 'Cappoquin'],
    'P56': ['Charleville'],
    'P61': ['Fermoy', 'Rathcormac'],
    'P67': ['Mitchelstown'],
    'P72': ['Bandon'],
    'P75': ['Bantry', 'Glengarriff'],
    'P81': ['Skibbereen', 'Baltimore'],
    'P85': ['Clonakilty', 'Rosscarbery'],
    'R14': ['Athy'],
    'R21': ['Bagenalstown'],
    'R32': ['Portlaoise', 'Abbeyleix', 'Portarlington'],
    'R35': ['Tullamore', 'Walsh Island, Offaly'],
    'R42': ['Birr', 'Banagher'],
    'R45': ['Edenderry'],
    'R51': ['Kildare'],
    'R56': ['Curragh'],
    'R93': ['Carlow', 'Tullow'],
    'R95': ['Kilkenny', 'Graiguenamanagh', 'Thomastown'],
    'T12': ['Cork city southside', 'Mahon', 'Passage West'],
    'T23': ['Cork city northside', 'Hollyhill'],
    'T34': ['Carrignavar'],
    'T45': ['Glanmire', 'Carrigtwohill'],
    'T56': ['Watergrasshill'],
    'V14': ['Shannon'],
    'V15': ['Kilrush', 'Kilkee'],
    'V23': ['Caherciveen'],
    'V31': ['Listowel'],
    'V35': ['Kilmallock'],
    'V42': ['Newcastle West'],
    'V92': ['Tralee', 'Dingle'],
    'V93': ['Kenmare', 'Killarney'],
    'V94': ['Limerick', 'Abbeyfeale', 'Newport', 'Killaloe-Ballina'],
    'V95': ['Miltown Malbay', 'Ennis', 'Kildysart'],
    'W12': ['Newbridge'],
    'W23': ['Maynooth', 'Celbridge', 'Leixlip'],
    'W34': ['Monasterevin'],
    'W91': ['Naas', 'Blessington'],
    'X35': ['Dungarvan'],
    'X42': ['Kilmacthomas', 'Bunmahon'],
    'X91': ['Waterford', 'Dunmore East', 'Tramore'],
    'Y14': ['Arklow'],
    'Y21': ['Enniscorthy', 'Bunclody'],
    'Y25': ['Gorey'],
    'Y34': ['New Ross', 'Fethard On Sea'],
    'Y35': ['Wexford', 'Rosslare'],
}

ROUTING_KEYS_COUNTY_MAP = {
    'A41': 'Dublin',
    'A42': 'Dublin',
    'A45': 'Dublin',
    'A63': 'Wicklow',
    'A67': 'Wicklow',
    'A75': 'Monaghan',
    'A81': 'Monaghan',
    'A82': 'Meath',
    'A83': 'Meath',
    'A84': 'Meath',
    'A85': 'Meath',
    'A86': 'Meath',
    'A91': 'Louth',
    'A92': 'Louth',
    'A94': 'Dublin',
    'A96': 'Dublin',
    'A98': 'Wicklow',
    'C15': 'Meath',
    'D01': 'Dublin',
    'D02': 'Dublin',
    'D03': 'Dublin',
    'D04': 'Dublin',
    'D05': 'Dublin',
    'D06': 'Dublin',
    'D6W': 'Dublin',
    'D07': 'Dublin',
    'D08': 'Dublin',
    'D09': 'Dublin',
    'D10': 'Dublin',
    'D11': 'Dublin',
    'D12': 'Dublin',
    'D13': 'Dublin',
    'D14': 'Dublin',
    'D15': 'Dublin',
    'D16': 'Dublin',
    'D17': 'Dublin',
    'D18': 'Dublin',
    'D20': 'Dublin',
    'D22': 'Dublin',
    'D24': 'Dublin',
    'E21': 'Tipperary',
    'E25': 'Tipperary',
    'E32': 'Tipperary',
    'E34': 'Tipperary',
    'E41': 'Tipperary',
    'E45': 'Tipperary',
    'E53': 'Tipperary',
    'E91': 'Tipperary',
    'F12': 'Mayo',
    'F23': 'Mayo',
    'F26': 'Down',
    'F28': 'Mayo',
    'F31': 'Mayo',
    'F35': 'Mayo',
    'F42': 'Roscommon',
    'F45': 'Roscommon',
    'F52': 'Roscommon',
    'F56': 'Sligo',
    'F91': 'Sligo',
    'F92': 'Donegal',
    'F93': 'Donegal',
    'F94': 'Donegal',
    'H12': 'Cavan',
    'H14': 'Cavan',
    'H16': 'Cavan',
    'H18': 'Monaghan',
    'H23': 'Monaghan',
    'H53': 'Galway',
    'H54': 'Galway',
    'H62': 'Galway',
    'H65': 'Galway',
    'H71': 'Galway',
    'H91': 'Galway',
    'K32': 'Dublin',
    'K34': 'Dublin',
    'K36': 'Dublin',
    'K45': 'Dublin',
    'K56': 'Dublin',
    'K67': 'Dublin',
    'K78': 'Dublin',
    'N37': 'Westmeath',
    'N39': 'Longford',
    'N41': 'Leitrim',
    'N91': 'Westmeath',
    'P12': 'Cork',
    'P14': 'Cork',
    'P17': 'Cork',
    'P24': 'Cork',
    'P25': 'Cork',
    'P31': 'Cork',
    'P32': 'Cork',
    'P36': 'Cork',
    'P43': 'Cork',
    'P47': 'Cork',
    'P51': 'Cork',
    'P56': 'Cork',
    'P61': 'Cork',
    'P67': 'Cork',
    'P72': 'Cork',
    'P75': 'Cork',
    'P81': 'Cork',
    'P85': 'Cork',
    'R14': 'Kildare',
    'R21': 'Carlow',
    'R32': 'Laois',
    'R35': 'Offaly',
    'R42': 'Offaly',
    'R45': 'Offaly',
    'R51': 'Kildare',
    'R56': 'Kildare',
    'R93': 'Carlow',
    'R95': 'Kilkenny',
    'T12': 'Cork',
    'T23': 'Cork',
    'T34': 'Cork',
    'T45': 'Cork',
    'T56': 'Cork',
    'V14': 'Clare',
    'V15': 'Clare',
    'V23': 'Kerry',
    'V31': 'Kerry',
    'V35': 'Limerick',
    'V42': 'Limerick',
    'V92': 'Kerry',
    'V93': 'Kerry',
    'V94': 'Limerick',
    'V95': 'Clare',
    'W12': 'Kildare',
    'W23': 'Kildare',
    'W34': 'Kildare',
    'W91': 'Kildare',
    'X35': 'Dungarvan',
    'X42': 'Waterford',
    'X91': 'Waterford',
    'Y14': 'Wicklow',
    'Y21': 'Wexford',
    'Y25': 'Wexford',
    'Y34': 'Wexford',
    'Y35': 'Wexford',
}



from difflib import SequenceMatcher
import urllib

from cached_property import cached_property
import requests

from eircode.eircode import Eircode
from eircode.constants import (
    IDENTITY_URL_PATH,
    EIRCODE_FINDER_URL_PATH
)
from eircode.proxy import Proxy

proxy_cli = Proxy(skip_setup=True)


class Addresses():

    def __init__(self):
        self._data = []

    def append(self, address):
        self._data.append(address)

    def ordered_best_fit(self, name):
        if isinstance(name, Address):
            name = name.display_name

        return sorted(
            [(i.match_score(name), i) for i in self._data],
            key=lambda i: i[0],
            reverse=True
        )


class Address():

    def __init__(self, address, **kwargs):
        '''

        :param address: The address you're searching for
        :kwargs display_name (optional): The display name found on eircode.ie
        :kwargs link (optional): The link that was used to get the info
        :kwargs eircode (optional): The eircode string
        '''
        self.input_address = address

        self.display_name = kwargs.get('display_name', None)
        self.link = kwargs.get('link', None)
        self._eircode = kwargs.get('eircode', None)

        self.proxy = kwargs.get('proxy', False)

        if not kwargs.get('skip_set', False):
            self.set(
                throw_ex=kwargs.get('throw_ex', False),
                reverse=kwargs.get('reverse', False),
            )

    def set(self, throw_ex=True, reverse=False):

        if self.proxy:

            self.proxy = True

            try:
                proxy_cli.setup()
            except:
                raise Exception('Could not set up proxy cli. Go to here for details: https://github.com/Ge0rg3/requests-ip-rotator')

            identity_response = proxy_cli.get(IDENTITY_URL_PATH).json()

            params = {
                'key': identity_response['key'],
                'address': self.input_address,
                'language': 'en',
                'geographicAddress': True,
                'clientVersion': None
            }

            finder_response = proxy_cli.get(
                EIRCODE_FINDER_URL_PATH + '?' + urllib.parse.urlencode(params)
            )

        else:
            identity_response = requests.get(IDENTITY_URL_PATH).json()
            params = {
                'key': identity_response['key'],
                'address': self.input_address,
                'language': 'en',
                'geographicAddress': True,
                'clientVersion': None
            }

            finder_response = requests.get(
                EIRCODE_FINDER_URL_PATH,
                params=params
            )

        finder_response = finder_response.json()

        if reverse:
            self._eircode = self.input_address
            self.display_name = ', '.join(finder_response['postalAddress'])
            return

        if 'postcode' in finder_response:
            self._eircode = finder_response['postcode']
            self.display_name = self.input_address
            return

        if finder_response.get('error', {}).get('code', None) == 403:
            if throw_ex:
                raise Exception(finder_response['error']['text'])
            else:
                print(f'Cannot search: {finder_response["error"]["text"]}')
                return

        options = finder_response['options']

        if options:
            addresses = Addresses()
            for option in options:
                addresses.append(
                    Address(
                        option['displayName'],
                        display_name=option['displayName'],
                        link=option['links'][0]['href'],
                        skip_set=True,
                        proxy=self.proxy
                    )
                )

            for address in addresses.ordered_best_fit(self.input_address):
                try:
                    self.display_name = address[1].eircode_data['display_name']
                    self._eircode = address[1].eircode_data['eircode']
                except ValueError:
                    pass
        else:
            print('TODO: use the autocomplete to see if we can get something')

    def match_score(self, name):
        if isinstance(name, Address):
            name = name.display_name

        return SequenceMatcher(
            None,
            self.display_name.lower(),
            name.lower()
        ).ratio()

    def serialize(self):
        return {
            'display_name': self.display_name,
            'link': self.link,
            'eircode': self.eircode.serialize()
        }

    @property
    def eircode(self):
        return Eircode(self._eircode)

    @cached_property
    def eircode_data(self):
        if self.proxy:
            data = proxy_cli.get(self.link).json()
        else:
            data = requests.get(self.link).json()

        data = requests.get(self.link).json()
        if data['options']:
            addresses = Addresses()
            for option in data['options']:
                addresses.append(
                    Address(
                        option['displayName'],
                        display_name=option['displayName'],
                        link=option['links'][0]['href'],
                        skip_set=True,
                        proxy=self.proxy
                    )
                )

            for address in addresses.ordered_best_fit(self.display_name):
                try:
                    if isinstance(address[1].eircode_data, str):
                        return {
                            'eircode': address[1].eircode_data,
                            'display_name': address[1].display_name
                        }
                    else:
                        return address[1].eircode_data
                except ValueError:
                    pass
        else:
            if data['result']['text'] == 'IncompleteAddressEntered':
                raise ValueError('Must go back')
            elif data['result']['text'] == 'NonUniqueAddress':
                raise ValueError('Must go back')
            else:
                return {
                    'eircode': data['postcode'],
                    'display_name': self.display_name
                }



from unittest import TestCase, skip

from eircode.proxy import Proxy


class ProxyTest(TestCase):

    @skip('if no aws creds just skip')
    def test_(self):
        proxy = Proxy(site='https://api.ipify.org')

        first_request = proxy.get('https://api.ipify.org')
        first = first_request.text

        proxy.shutdown()
        proxy.setup(force=True)

        second_request = proxy.get('https://api.ipify.org')
        second = second_request.text

        self.assertNotEqual(first, second)



import urllib
import urllib.parse

import requests
from requests_ip_rotator import ApiGateway


class Proxy():

    def __init__(self, skip_setup=False, site=None):
        self.gateway = None
        self.session = None

        self.site = site if site is not None else 'https://api-finder.eircode.ie'
        self.hostname = urllib.parse.urlparse(self.site).netloc

        self._is_setup = False

        if not skip_setup:
            self.setup()

    def setup(self, force=False):
        if not force and self._is_setup:
            return

        self.gateway = ApiGateway(
            self.site,
            regions=['eu-west-1']
        )
        self.gateway.start()

        self.session = requests.Session()
        self.session.mount(
            self.site,
            self.gateway
        )

        self._is_setup = True

    def get(self, url, params={}):
        try:
            return self.session.get(
                url + '?' + urllib.parse.urlencode(params)
            )
        except:
            self.shutdown()
            self.setup(force=True)
            return self.get(url, params=params)

    def shutdown(self):
        self.gateway.shutdown()



from eircode.constants import ROUTING_KEYS_TOWNS_MAP, ROUTING_KEYS_COUNTY_MAP


class Eircode():

    def __init__(self, eircode):
        self.eircode = eircode

    def __repr__(self):
        return f'Eircode({self.eircode})'

    @property
    def routing_key(self):
        if not self.eircode:
            return None
        return self.eircode[:3]

    @property
    def unique_identifier(self):
        if not self.eircode:
            return None
        return self.eircode[3:]

    @property
    def county(self):
        if not self.eircode:
            return None
        return ROUTING_KEYS_COUNTY_MAP[self.routing_key]

    @property
    def towns(self):
        if not self.eircode:
            return None
        return ROUTING_KEYS_TOWNS_MAP[self.routing_key]

    def serialize(self):
        return {
            'routing_key': self.routing_key,
            'unique_identifier': self.unique_identifier,
            'eircode': self.eircode,
            'county': self.county,
            'towns': self.towns
        }



from unittest import TestCase

from eircode.address import Address, Addresses


class AddressesTest(TestCase):

    def test_append(self):
        addresses = Addresses()
        addresses.append(Address('', display_name='something', skip_set=True))

        self.assertEqual(len(addresses._data), 1)

    def test_ordered_best_fit(self):
        addresses = Addresses()
        addresses.append(Address('', display_name='something', skip_set=True))
        addresses.append(Address('', display_name='something else', skip_set=True))
        addresses.append(Address('', display_name='blah something', skip_set=True))

        self.assertEqual(
            addresses.ordered_best_fit('blahblah')[0][1].display_name,
            'blah something'
        )


class AddressTest(TestCase):

    def test_match_score(self):
        self.assertEqual(
            Address('', display_name='something', skip_set=True).match_score(
                Address('', display_name='something', skip_set=True)
            ),
            1
        )

        self.assertEqual(
            Address('', display_name='something', skip_set=True).match_score(
                Address('', display_name='aaaaaa', skip_set=True)
            ),
            0
        )

        self.assertGreater(
            Address('', display_name='something', skip_set=True).match_score(
                Address('', display_name='some', skip_set=True)
            ),
            0.5
        )

    def test_eircode(self):
        self.assertEqual(
            Address(
                '',
                eircode='D14AAAA',
                skip_set=True
            ).eircode.serialize(),
            {
                'county': 'Dublin',
                'eircode': 'D14AAAA',
                'routing_key': 'D14',
                'towns': ['Dublin 14'],
                'unique_identifier': 'AAAA'
            }
        )

    def test_serialize(self):
        self.assertEqual(
            Address(
                '',
                display_name='something',
                eircode='D14AAAA',
                skip_set=True
            ).serialize(),
            {
                'display_name': 'something',
                'eircode': {
                    'county': 'Dublin',
                    'eircode': 'D14AAAA',
                    'routing_key': 'D14',
                    'towns': ['Dublin 14'],
                    'unique_identifier': 'AAAA'
                },
                'link': None}
        )



from setuptools import (
    find_packages,
    setup
)

INSTALL_REQUIRES = (
    'requests',
    'cached_property',
    'requests-ip-rotator'
)

setup(
    name='eircode',
    version='0.0.15',
    python_requires='>=3.5',
    author='Robert Lucey',
    url='https://github.com/RobertLucey/eircode',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=INSTALL_REQUIRES
)



from unittest import TestCase

from eircode.eircode import Eircode


class EircodeTest(TestCase):

    def test_repr(self):
        self.assertEqual(
            str(Eircode('D14AAAA')),
            'Eircode(D14AAAA)'
        )

    def test_routing_key(self):
        self.assertEqual(
            Eircode('D14AAAA').routing_key,
            'D14'
        )

    def test_unique_identifier(self):
        self.assertEqual(
            Eircode('D14AAAA').unique_identifier,
            'AAAA'
        )

    def test_county(self):
        self.assertEqual(
            Eircode('D14AAAA').county,
            'Dublin'
        )

    def test_towns(self):
        self.assertEqual(
            Eircode('D14AAAA').towns,
            ['Dublin 14']
        )

    def test_serialize(self):
        self.assertEqual(
            Eircode('D14AAAA').serialize(),
            {
                'county': 'Dublin',
                'eircode': 'D14AAAA',
                'routing_key': 'D14',
                'towns': ['Dublin 14'],
                'unique_identifier': 'AAAA'
            }
        )
